@@ -27,6 +27,7 @@
 #include "RDocumentInterface.h"
 #include "RBlock.h"
 #include "RBlockListener.h"
+#include "RBlockReferenceEntity.h"
 #include "RCoordinateEvent.h"
 #include "RCoordinateListener.h"
 #include "RDimStyle.h"
@@ -87,6 +88,13 @@ RDocumentInterface::RDocumentInterface(RDocument& document)
     previewDocument(NULL) {
     //pressEvent(NULL) {
 
+    QObject::connect(
+        &fileSystemWatcher, &QFileSystemWatcher::fileChanged,
+        [this](const QString& path) {
+            xRefFileChanged(path);
+        }
+    );
+
     RDebug::incCounter("RDocumentInterface");
 }
 
@@ -528,6 +536,17 @@ void RDocumentInterface::suspend() {
 }
 
 void RDocumentInterface::resume() {
+    // ask user if changed XRefs should be reloaded:
+    if (!dirtyXRefPaths.isEmpty()) {
+        RMainWindow* appWin = RMainWindow::getMainWindow();
+        if (appWin!=NULL) {
+            // do the actual reloading in the GUI:
+            appWin->reloadXRefs(this, dirtyXRefPaths);
+        }
+
+        dirtyXRefPaths.clear();
+    }
+
     if (currentSnap!=NULL) {
         currentSnap->showUiOptions();
     }
@@ -1335,6 +1354,8 @@ RDocumentInterface::IoErrorCode RDocumentInterface::importFile(
 
     delete fileImporter;
 
+    loadXRefs();
+
     if (mainWindow!=NULL && notify==true && notifyGlobalListeners==true) {
         mainWindow->notifyListeners();
         mainWindow->notifyImportListenersPost(this);
@@ -1439,6 +1460,7 @@ void RDocumentInterface::undoToTag(const QString& tag) {
  * Transaction based undo.
  */
 void RDocumentInterface::undo() {
+    //qDebug() << "RDocumentInterface::undo: vvvvvvvvvvv";
     RMainWindow* mainWindow = RMainWindow::getMainWindow();
     clearPreview();
 
@@ -1460,12 +1482,14 @@ void RDocumentInterface::undo() {
             mainWindow->handleUserMessage(QString("Undo:") + " " + t[i].getText());
         }
     }
+    //qDebug() << "RDocumentInterface::undo: ^^^^^^^^^^";
 }
 
 /**
  * Transaction based redo.
  */
 void RDocumentInterface::redo() {
+    //qDebug() << "RDocumentInterface::redo: vvvvvvvvvvv";
     RMainWindow* mainWindow = RMainWindow::getMainWindow();
     clearPreview();
 
@@ -1488,6 +1512,7 @@ void RDocumentInterface::redo() {
             mainWindow->handleUserMessage(QString("Redo:") + " " + t[i].getText());
         }
     }
+    //qDebug() << "RDocumentInterface::redo: ^^^^^^^^^^";
 }
 
 /**
@@ -2327,6 +2352,7 @@ void RDocumentInterface::objectChangeEvent(RTransaction& transaction) {
     //bool layerHasChanged = false;
     QSet<RLayer::Id> changedLayerIds;
     bool blockHasChanged = false;
+    bool blockXRefHasChanged = false;
     bool layoutHasChanged = false;
     bool viewHasChanged = false;
     bool entityHasChanged = false;
@@ -2456,18 +2482,34 @@ void RDocumentInterface::objectChangeEvent(RTransaction& transaction) {
         if (type==RS::ObjectBlock) {
             QSharedPointer<RBlock> block = object.dynamicCast<RBlock> ();
             if (!block.isNull()) {
-                //if (block->getId()!=document.getModelSpaceBlockId()) {
+
                 if (block->getId()!=document.getCurrentBlockId()) {
                     blockHasChanged = true;
-                    //document.queryBlockReferences(block->getId());
                 }
-                else {
-                    QList<RPropertyChange> propertyChanges = transaction.getPropertyChanges(objectId);
-                    for (int i=0; i<propertyChanges.length(); i++) {
-                        if (propertyChanges[i].getPropertyTypeId()==RBlock::PropertyName) {
-                            // current block renamed (needs block list update):
-                            blockHasChanged = true;
-                        }
+
+                bool blockNameHasChanged = false;
+                QString oldBlockName = "";
+                QString newBlockName = "";
+
+                bool xRefFileNameHasChanged = false;
+                QString oldXRefFileName = "";
+                QString newXRefFileName = "";
+
+                QList<RPropertyChange> propertyChanges = transaction.getPropertyChanges(objectId);
+                for (int i=0; i<propertyChanges.length(); i++) {
+                    if (propertyChanges[i].getPropertyTypeId()==RBlock::PropertyName) {
+                        // current block renamed (needs block list update):
+                        blockHasChanged = true;
+                        blockNameHasChanged = true;
+                        newBlockName = propertyChanges[i].getNewValue().toString();
+                        oldBlockName = propertyChanges[i].getOldValue().toString();
+                    }
+
+                    if (propertyChanges[i].getPropertyTypeId()==RBlock::PropertyXRefFileName) {
+                        // handle undo / redo for XRef file name changes
+                        xRefFileNameHasChanged = true;
+                        newXRefFileName = propertyChanges[i].getNewValue().toString();
+                        oldXRefFileName = propertyChanges[i].getOldValue().toString();
                     }
                 }
 
@@ -2476,6 +2518,63 @@ void RDocumentInterface::objectChangeEvent(RTransaction& transaction) {
                     QSet<RObject::Id> ids = document.queryBlockReferences(*it);
                     deselectEntities(ids);
                 }
+
+                bool binding = !oldXRefFileName.isEmpty() && newXRefFileName.isEmpty();
+
+                // block is XRef or was XRef (binding):
+                if (block->isXRef() || binding) {
+                    blockXRefHasChanged = true;
+
+                    if (block->isUndone()) {
+                        block->unloadXRef();
+                        fileSystemWatcher.removePath(block->getFullXRefFilePath());
+                            //.addPath(block->getFullXRefFilePath());
+                    }
+                    else {
+                        if (blockNameHasChanged) {
+                            if (transaction.isType(RTransaction::Undo)) {
+                                // unload block with new name while undoing:
+                                block->setName(newBlockName);
+                                block->unloadXRef();
+                                block->setName(oldBlockName);
+                            }
+                            else if (transaction.isType(RTransaction::Redo)) {
+                                // unload block with old name while redoing:
+                                block->setName(oldBlockName);
+                                block->unloadXRef();
+                                block->setName(newBlockName);
+                            }
+                            else {
+                                // unload block with old name while renaming:
+                                block->setName(oldBlockName);
+                                block->unloadXRef();
+                                block->setName(newBlockName);
+                            }
+                        }
+                        if (xRefFileNameHasChanged) {
+                            if (transaction.isType(RTransaction::Redo)) {
+                                if (!oldXRefFileName.isEmpty() && newXRefFileName.isEmpty()) {
+                                    // redo binding:
+                                    block->setXRefFileName(oldXRefFileName);
+                                    block->bindXRef(false);
+                                    block->setXRefFileName(newXRefFileName);
+                                }
+                            }
+                        }
+
+                        if (!binding) {
+                            // force XRef reload:
+                            block->setXRefLoaded(false);
+
+                            QStringList xRefFileNames;
+                            RBlockProxy* proxy = RBlock::getBlockProxy();
+                            proxy->loadXRef(block.data(), xRefFileNames);
+
+                            fileSystemWatcher.addPaths(xRefFileNames);
+                            fileSystemWatcher.addPath(block->getFullXRefFilePath());
+                        }
+                    }
+                }
                 continue;
             }
         }
@@ -2492,22 +2591,29 @@ void RDocumentInterface::objectChangeEvent(RTransaction& transaction) {
     }
 
     // notify global listeners if this is not the clipboard document interface:
-    if (RMainWindow::hasMainWindow() && notifyGlobalListeners) {
+    RMainWindow* mainWindow = RMainWindow::getMainWindow();
+    if (mainWindow!=NULL && notifyGlobalListeners) {
         if (ucsHasChanged) {
-            RMainWindow::getMainWindow()->notifyUcsListeners(this);
+            mainWindow->notifyUcsListeners(this);
         }
         if (linetypeHasChanged) {
             // TODO:
             //RMainWindow::getMainWindow()->notifyLinetypeListeners(this);
         }
         if (!changedLayerIds.isEmpty()) {
-            RMainWindow::getMainWindow()->notifyLayerListeners(this, changedLayerIdList);
+            mainWindow->notifyLayerListeners(this, changedLayerIdList);
         }
         if (blockHasChanged || layoutHasChanged) {
-            RMainWindow::getMainWindow()->notifyBlockListeners(this);
+            mainWindow->notifyBlockListeners(this);
+        }
+        if (blockXRefHasChanged) {
+            mainWindow->notifyBlockListenersXRef();
+            QList<RLayer::Id> layerIds;
+            // update all layers:
+            mainWindow->notifyLayerListeners(this, layerIds);
         }
         if (viewHasChanged) {
-            RMainWindow::getMainWindow()->notifyViewListeners(this);
+            mainWindow->notifyViewListeners(this);
         }
     }
 
@@ -2798,3 +2904,43 @@ QVariant RDocumentInterface::eval(const QString& ext, const QString& script) {
     }
     return handler->eval(script);
 }
+
+void RDocumentInterface::xRefFileChanged(const QString& dirtyXRefPath) {
+    //loadXRefs(path);
+
+    // remember which XRefs have been changed to ask user to reload on resume:
+    dirtyXRefPaths.insert(dirtyXRefPath);
+}
+
+void RDocumentInterface::loadXRefs(const QSet<QString>& paths) {
+    QStringList xRefFileNames;
+
+    RBlockProxy* proxy = RBlock::getBlockProxy();
+    if (proxy==NULL) {
+        return;
+    }
+
+    RTransaction t = proxy->loadXRefs(document, paths, xRefFileNames);
+
+    objectChangeEvent(t);
+
+    if (xRefFileNames.isEmpty()) {
+        return;
+    }
+
+    fileSystemWatcher.addPaths(xRefFileNames);
+}
+
+void RDocumentInterface::bindXRef(RBlock* block) {
+    if (block==NULL) {
+        return;
+    }
+
+    RBlockProxy* proxy = RBlock::getBlockProxy();
+    if (proxy==NULL) {
+        return;
+    }
+
+    RTransaction t = proxy->bindXRef(block);
+    objectChangeEvent(t);
+}
