@@ -270,7 +270,7 @@ bool RPolyline::appendShape(const RShape& shape, bool prepend) {
 
     RVector connectionPoint;
     RVector nextPoint;
-    double gap;
+    RVector plPoint;
     if (prepend) {
         // prepend:
         connectionPoint = shape.getEndPoint();
@@ -279,7 +279,8 @@ bool RPolyline::appendShape(const RShape& shape, bool prepend) {
             // first point:
             appendVertex(connectionPoint);
         }
-        gap = vertices.first().getDistanceTo(connectionPoint);
+        plPoint = vertices.first();
+        //gap = vertices.first().getDistanceTo(connectionPoint);
     }
     else {
         // append:
@@ -289,12 +290,14 @@ bool RPolyline::appendShape(const RShape& shape, bool prepend) {
             // first point:
             appendVertex(connectionPoint);
         }
-        gap = vertices.last().getDistanceTo(connectionPoint);
+        plPoint = vertices.last();
+        //gap = vertices.last().getDistanceTo(connectionPoint);
     }
 
+    double gap = plPoint.getDistanceTo(connectionPoint);
     if (!RMath::fuzzyCompare(gap, 0.0, 1.0e-3)) {
         qWarning() << "RPolyline::appendShape: "
-                   << "arc or line not connected to polyline at " << connectionPoint << ":"
+                   << "arc or line not connected to polyline at " << plPoint << ":"
                    << "\nshape:" << shape
                    << "\ngap: " << gap;
         ret = false;
@@ -1177,6 +1180,17 @@ QList<RPolyline> RPolyline::getOutline() const {
     }
 }
 
+QList<QSharedPointer<RShape>> RPolyline::getSegments() const {
+    QList<QSharedPointer<RShape>> ret;
+    for (int i=0; i<countSegments(); i++) {
+        QSharedPointer<RShape> seg = getSegmentAt(i);
+        if (!seg.isNull()) {
+            ret.append(seg);
+        }
+    }
+    return ret;
+}
+
 /**
  * \return Number of segments. The number of segments equals the
  *      number of vertices for a closed polyline and one less for
@@ -1299,9 +1313,18 @@ bool RPolyline::contains(const RVector& point, bool borderIsInside, double toler
     }
 
     if (hasArcSegments()) {
-        // TODO: not always reliable:
-        QPainterPath pp = toPainterPath();
-        return pp.contains(QPointF(point.x, point.y));
+        if (RPolyline::hasProxy()) {
+            bool ret = RPolyline::getPolylineProxy()->contains(*this, point, borderIsInside, tolerance);
+            qDebug("pl:");
+            dump();
+            qDebug() << "point: " << point.x << " / " << point.y;
+            return ret;
+        }
+        else {
+            // not reliable, false negatives:
+            QPainterPath pp = toPainterPath();
+            return pp.contains(QPointF(point.x, point.y));
+        }
     }
 
     int nvert = vertices.size();
@@ -1316,6 +1339,7 @@ bool RPolyline::contains(const RVector& point, bool borderIsInside, double toler
     return c;
 }
 
+
 /**
  * Checks if the given shape is completely inside this closed polygon. If this
  * polyline is not closed, false is returned.
@@ -1369,13 +1393,14 @@ bool RPolyline::containsShape(const RShape& shape) const {
             const RCircle& circle = dynamic_cast<const RCircle&>(shape);
             RVector p1 = circle.getCenter() + RVector(circle.getRadius(), 0);
             RVector p2 = circle.getCenter() + RVector(-circle.getRadius(), 0);
-            if (contains(p1) || contains(p2)) {
+            if (contains(p1) && contains(p2)) {
                 return true;
             }
             return false;
         }
         else {
             // other shapes:
+            // shape is inside if one point on the shape is inside and there are no intersections (see above):
             RVector pointOnShape = shape.getPointOnShape();
             if (contains(pointOnShape, true)) {
                 return true;
